%% Derman-Kani-Chriss implied volatility tree.
% Computes an implied trinomial tree using market implied volatilities on
% Bank of America.  Constructs an IV surface using interpolation or
% regression and computes market prices of calls and puts based on the IV
% surface.  Shows that the tree fully recovers the market prices of calls
% and puts.

% By Fabrice Douglas Rouah.

clc; clear;

% Input the spot price, time (T), risk free rate (r), initial ATM
% volatility (v), number of time steps (n), and skew
Spot = 50;
T  = 5;
rf = 0.05;
v  = 0.15;
n  = 6;
skew = 0.05;

% Compute the time increment (dt) and up and down stock price 
% moves (u and d)
dt = T/n;
u  = exp(v*sqrt(2*dt));
d  = 1/u;

% Inline function for the Black Scholes call
BSCall = inline('s*exp(-q*T)*normcdf((log(s/K) + (r-q+v^2/2)*T)/v/sqrt(T)) - K*exp(-r*T)*normcdf((log(s/K) + (r-q+v^2/2)*T)/v/sqrt(T) - v*sqrt(T))',...
	     's','K','r','q','v','T');

% Inline function for the Black Scholes put
BSPut = inline('K*exp(-r*T)*normcdf(-(log(s/K) + (r-q+v^2/2)*T)/v/sqrt(T) + v*sqrt(T)) - s*exp(-q*T)*normcdf(-(log(s/K) + (r-q+v^2/2)*T)/v/sqrt(T))',...
	    's','K','r','q','v','T');

%% Stock Price Dynamics
S = zeros(2*n+1,n+1);
S(n+1,1) = Spot;
J = [n:-1:-n];
for j = 2:n+1
    for i = n-j+2:n+j
        S(i,j) = Spot*u^J(i);
	end
end

% Calculate Implied Volatility Structure.  
% Example from Derman, Kani and Chriss
IV = zeros(2*n+1,n+1);
for j = 1:n+1
    for i = n-j+2:n+j;
        IV(i,j) = v - skew*(S(i,j)-Spot)/100;
	end
end

%% Create market calls and puts from implied vol surface
Call = zeros(2*n+1,n+1);
Put  = zeros(2*n+1,n+1);
for j=1:n+1
	for i=n-j+2:n+j
		Put(i,j)  =  BSPut(Spot,S(i,j),rf,0,IV(i,j),j*dt);
		Call(i,j) = BSCall(Spot,S(i,j),rf,0,IV(i,j),j*dt);
	end
end

%% Create the Local Volatility surface, 
% and the "up", "middle", and "down" probabilities.
[LocalVol Q pu pm pd] = DKCTrinomial3(Call,Put,Spot,T(end),rf,v,n);


%% Recover the calls and puts
TreePut  = zeros(2*n-1,n);
TreeCall = zeros(2*n-1,n);

for t=1:n
	PU = pu(n-t+1:n+t-1, 1:t);
	PM = pm(n-t+1:n+t-1, 1:t);
	PD = pd(n-t+1:n+t-1, 1:t);
    LOCALVOL = LocalVol(n-t+1:n+t-1, 1:t);
	SS =  S(n-t+1:n+t+1, 1:t+1);
	CALL = Call(n-t+1:n+t+1, 1:t+1);
	PUT  =  Put(n-t+1:n+t+1, 1:t+1);
	for k=2:2*t
		Strike = SS(k,t);
		C(:,t+1) = max(SS(:,t+1) - Strike, 0);
		P(:,t+1) = max(Strike - SS(:,t+1), 0);
		for j = t:-1:1
			for i = t-j+2:t+j
 				P(i,j) = exp(-rf*dt)*(PU(i-1,j)*P(i-1,j+1) + PM(i-1,j)*P(i,j+1) + PD(i-1,j)*P(i+1,j+1));
				C(i,j) = exp(-rf*dt)*(PU(i-1,j)*C(i-1,j+1) + PM(i-1,j)*C(i,j+1) + PD(i-1,j)*C(i+1,j+1));
			end
		end
		TreePut(n-t+k-1,t)  = P(t+1,1);
		TreeCall(n-t+k-1,t) = C(t+1,1);
		clear C P
	end
end


%% Calculate the errors between the market prices of calls and puts, and the
% recovered prices from the IV tree.

disp('-----------Market Prices of puts --------------------------')
Put
disp('-----------Recovered prices of puts and error -------------')
TreePut
Put_SSE  = sum(sum((TreePut  -  Put(2:2*n,1:n)).^2))

disp('-----------Market Prices of calls--------------------------')
Call
disp('-----------Recovered prices of calls and error ------------')
TreeCall
Call_SSE = sum(sum((TreeCall - Call(2:2*n,1:n)).^2))

